@@ -1,47 +1,69 @@
 """IPython terminal interface using prompt_toolkit in place of readline"""
 from __future__ import print_function
 
+import asyncio
 import base64
 import errno
 from getpass import getpass
 from io import BytesIO
 import os
+from queue import Empty
 import signal
 import subprocess
 import sys
+from tempfile import TemporaryDirectory
 import time
 from warnings import warn
 
-try:
-    from queue import Empty  # Py 3
-except ImportError:
-    from Queue import Empty  # Py 2
+from typing import Dict as DictType, Any as AnyType
 
 from zmq import ZMQError
 from IPython.core import page
-from IPython.utils.py3compat import cast_unicode_py2, input
-from ipython_genutils.tempdir import NamedFileInTemporaryDirectory
-from traitlets import (Bool, Integer, Float, Unicode, List, Dict, Enum,
-                       Instance, Any)
+from traitlets import (
+    Bool,
+    Integer,
+    Float,
+    Unicode,
+    List,
+    Dict,
+    Enum,
+    Instance,
+    Any,
+)
 from traitlets.config import SingletonConfigurable
 
 from .completer import ZMQCompleter
 from .zmqhistory import ZMQHistoryManager
 from . import __version__
 
+# Discriminate version3 for asyncio
+from prompt_toolkit import __version__ as ptk_version
+PTK3 = ptk_version.startswith('3.')
+
+if not PTK3:
+    # use_ayncio_event_loop obsolete in PKT3
+    from prompt_toolkit.eventloop.defaults import use_asyncio_event_loop
+
 from prompt_toolkit.completion import Completer, Completion
 from prompt_toolkit.document import Document
 from prompt_toolkit.enums import DEFAULT_BUFFER, EditingMode
-from prompt_toolkit.filters import (Condition, has_focus, has_selection,
-                                    vi_insert_mode, emacs_insert_mode, is_done)
+from prompt_toolkit.filters import (
+    Condition,
+    has_focus,
+    has_selection,
+    vi_insert_mode,
+    emacs_insert_mode,
+    is_done,
+)
 from prompt_toolkit.history import InMemoryHistory
-from prompt_toolkit.patch_stdout import patch_stdout
 from prompt_toolkit.shortcuts.prompt import PromptSession
-from prompt_toolkit.shortcuts import print_formatted_text
+from prompt_toolkit.shortcuts import print_formatted_text, CompleteStyle
 from prompt_toolkit.key_binding import KeyBindings
 from prompt_toolkit.lexers import PygmentsLexer
-from prompt_toolkit.layout.processors import (ConditionalProcessor,
-                                              HighlightMatchingBracketProcessor)
+from prompt_toolkit.layout.processors import (
+    ConditionalProcessor,
+    HighlightMatchingBracketProcessor,
+)
 from prompt_toolkit.styles import merge_styles
 from prompt_toolkit.styles.pygments import (style_from_pygments_cls,
                                             style_from_pygments_dict)
@@ -54,6 +76,8 @@ from pygments.lexers import get_lexer_by_name
 from pygments.util import ClassNotFound
 from pygments.token import Token
 
+from jupyter_client.utils import run_sync
+
 
 def ask_yes_no(prompt, default=None, interrupt=None):
     """Asks a question and returns a boolean (y/n) answer.
@@ -88,6 +112,15 @@ def ask_yes_no(prompt, default=None, interrupt=None):
     return answers[ans]
 
 
+async def async_input(prompt, loop=None):
+    """Simple async version of input using a the default executor"""
+    if loop is None:
+        loop = asyncio.get_event_loop()
+
+    raw = await loop.run_in_executor(None, input, prompt)
+    return raw
+
+
 def get_pygments_lexer(name):
     name = name.lower()
     if name == 'ipython2':
@@ -118,9 +151,26 @@ class JupyterPTCompleter(Completer):
             code=document.text,
             cursor_pos=document.cursor_position
         )
-        start_pos = content['cursor_start'] - document.cursor_position
-        for m in content['matches']:
-            yield Completion(m, start_position=start_pos)
+        meta = content["metadata"]
+
+        if "_jupyter_types_experimental" in meta:
+            try:
+                new_meta = {}
+                for c, m in zip(
+                    content["matches"], meta["_jupyter_types_experimental"]
+                ):
+                    new_meta[c] = m["type"]
+                meta = new_meta
+            except Exception:
+                pass
+
+        start_pos = content["cursor_start"] - document.cursor_position
+        for m in content["matches"]:
+            yield Completion(
+                m,
+                start_position=start_pos,
+                display_meta=meta.get(m, "?"),
+            )
 
 
 class ZMQTerminalInteractiveShell(SingletonConfigurable):
@@ -211,8 +261,9 @@ class ZMQTerminalInteractiveShell(SingletonConfigurable):
         """
     )
 
-    callable_image_handler = Any(config=True, help=
-        """
+    callable_image_handler = Any(
+        config=True,
+        help="""
         Callable object called via 'callable' image handler with one
         argument, `data`, which is `msg["content"]["data"]` where
         `msg` is the message from iopub channel.  For example, you can
@@ -253,15 +304,29 @@ class ZMQTerminalInteractiveShell(SingletonConfigurable):
         help="""Set to display confirmation dialog on exit.
         You can always use 'exit' or 'quit', to force a
         direct exit without any confirmation.
-        """
+        """,
     )
 
-    highlight_matching_brackets = Bool(True,
-        help="Highlight matching brackets.",
+    display_completions = Enum(
+        ("column", "multicolumn", "readlinelike"),
+        help=(
+            "Options for displaying tab completions, 'column', 'multicolumn', and "
+            "'readlinelike'. These options are for `prompt_toolkit`, see "
+            "`prompt_toolkit` documentation for more information."
+        ),
+        default_value="multicolumn",
+    ).tag(config=True)
+    
+    prompt_includes_vi_mode = Bool(True,
+        help="Display the current vi mode (when using vi editing mode)."
     ).tag(config=True)
 
-    manager = Instance('jupyter_client.KernelManager', allow_none=True)
-    client = Instance('jupyter_client.KernelClient', allow_none=True)
+    highlight_matching_brackets = Bool(True, help="Highlight matching brackets.",).tag(
+        config=True
+    )
+
+    manager = Instance("jupyter_client.KernelManager", allow_none=True)
+    client = Instance("jupyter_client.KernelClient", allow_none=True)
 
     def _client_changed(self, name, old, new):
         self.session_id = new.session.session
@@ -304,16 +369,25 @@ class ZMQTerminalInteractiveShell(SingletonConfigurable):
         self.history_manager = ZMQHistoryManager(client=self.client)
         self.configurables.append(self.history_manager)
 
-    def get_prompt_tokens(self):
+    def vi_mode(self):
+        if (getattr(self, 'editing_mode', None) == 'vi'
+                and self.prompt_includes_vi_mode):
+            return '['+str(self.pt_cli.app.vi_state.input_mode)[3:6]+'] '
+        return ''
+    
+    def get_prompt_tokens(self, ec=None):
+        if ec is None:
+            ec = self.execution_count
         return [
+            (Token.Prompt, self.vi_mode()),
             (Token.Prompt, 'In ['),
-            (Token.PromptNum, str(self.execution_count)),
+            (Token.PromptNum, str(ec)),
             (Token.Prompt, ']: '),
         ]
 
     def get_continuation_tokens(self, width):
         return [
-            (Token.Prompt, (' ' * (width - 2)) + ': '),
+            (Token.Prompt, (" " * (width - 5)) + "...: "),
         ]
 
     def get_out_prompt_tokens(self):
@@ -328,7 +402,26 @@ class ZMQTerminalInteractiveShell(SingletonConfigurable):
         print_formatted_text(PygmentsTokens(tokens), end='',
                              style = self.pt_cli.app.style)
 
-    kernel_info = {}
+    def get_remote_prompt_tokens(self):
+        return [
+            (Token.RemotePrompt, self.other_output_prefix),
+        ]
+
+    def print_remote_prompt(self, ec=None):
+        tokens = self.get_remote_prompt_tokens() + self.get_prompt_tokens(ec=ec)
+        print_formatted_text(
+            PygmentsTokens(tokens), end="", style=self.pt_cli.app.style
+        )
+
+    @property
+    def pt_complete_style(self):
+        return {
+            "multicolumn": CompleteStyle.MULTI_COLUMN,
+            "column": CompleteStyle.COLUMN,
+            "readlinelike": CompleteStyle.READLINE_LIKE,
+        }[self.display_completions]
+
+    kernel_info: DictType[str, AnyType] = {}
 
     def init_kernel_info(self):
         """Wait for a kernel to be ready, and store kernel info"""
@@ -339,9 +432,9 @@ class ZMQTerminalInteractiveShell(SingletonConfigurable):
         while True:
             try:
                 reply = self.client.get_shell_msg(timeout=1)
-            except Empty:
+            except Empty as e:
                 if (time.time() - tic) > timeout:
-                    raise RuntimeError("Kernel didn't respond to kernel_info_request")
+                    raise RuntimeError("Kernel didn't respond to kernel_info_request") from e
             else:
                 if reply['parent_header'].get('msg_id') == msg_id:
                     self.kernel_info = reply['content']
@@ -349,14 +442,16 @@ class ZMQTerminalInteractiveShell(SingletonConfigurable):
 
     def show_banner(self):
         print(self.banner.format(version=__version__,
-                         kernel_banner=self.kernel_info.get('banner', '')))
+                         kernel_banner=self.kernel_info.get('banner', '')),end='',flush=True)
 
     def init_prompt_toolkit_cli(self):
         if self.simple_prompt or ('JUPYTER_CONSOLE_TEST' in os.environ):
             # Simple restricted interface for tests so we can find prompts with
             # pexpect. Multi-line input not supported.
-            def prompt():
-                return cast_unicode_py2(input('In [%d]: ' % self.execution_count))
+            async def prompt():
+                prompt = 'In [%d]: ' % self.execution_count
+                raw = await async_input(prompt)
+                return raw
             self.prompt_for_code = prompt
             self.print_out_prompt = \
                 lambda: print('Out[%d]: ' % self.execution_count, end='')
@@ -400,13 +495,17 @@ class ZMQTerminalInteractiveShell(SingletonConfigurable):
         def _(event):
             event.cli.suspend_to_background()
 
+        @kb.add("c-o", filter=(has_focus(DEFAULT_BUFFER) & emacs_insert_mode))
+        def _(event):
+            event.current_buffer.insert_text("\n")
+
         # Pre-populate history from IPython's history database
         history = InMemoryHistory()
         last_cell = u""
         for _, _, cell in self.history_manager.get_tail(self.history_load_length,
                                                         include_latest=True):
             # Ignore blank lines and consecutive duplicates
-            cell = cast_unicode_py2(cell.rstrip())
+            cell = cell.rstrip()
             if cell and (cell != last_cell):
                 history.append_string(cell)
 
@@ -415,6 +514,7 @@ class ZMQTerminalInteractiveShell(SingletonConfigurable):
             Token.PromptNum: '#00ff00 bold',
             Token.OutPrompt: '#ff2200',
             Token.OutPromptNum: '#ff0000 bold',
+            Token.RemotePrompt: '#999900',
         }
         if self.highlighting_style:
             style_cls = get_style_by_name(self.highlighting_style)
@@ -449,14 +549,21 @@ class ZMQTerminalInteractiveShell(SingletonConfigurable):
             Condition(lambda: self.highlight_matching_brackets))
         ]
 
+        # Tell prompt_toolkit to use the asyncio event loop.
+        # Obsolete in prompt_toolkit.v3
+        if not PTK3:
+            use_asyncio_event_loop()
+
         self.pt_cli = PromptSession(
             message=(lambda: PygmentsTokens(self.get_prompt_tokens())),
             multiline=True,
+            complete_style=self.pt_complete_style,
             editing_mode=editing_mode,
             lexer=PygmentsLexer(get_pygments_lexer(lexer)),
             prompt_continuation=(
-                lambda width, lineno, is_soft_wrap:
-                PygmentsTokens(self.get_continuation_tokens(width))
+                lambda width, lineno, is_soft_wrap: PygmentsTokens(
+                    self.get_continuation_tokens(width)
+                )
             ),
             key_bindings=kb,
             history=history,
@@ -465,21 +572,20 @@ class ZMQTerminalInteractiveShell(SingletonConfigurable):
             style=style,
             input_processors=input_processors,
             color_depth=(ColorDepth.TRUE_COLOR if self.true_color else None),
-
         )
 
-    def prompt_for_code(self):
+    async def prompt_for_code(self):
         if self.next_input:
             default = self.next_input
             self.next_input = None
         else:
             default = ''
 
-        with patch_stdout(raw=True):
-            text = self.pt_cli.prompt(
-                default=default,
-#                pre_run=self.pre_prompt,# reset_current_buffer=True,
-            )
+        if PTK3:
+            text = await self.pt_cli.prompt_async(default=default)
+        else:
+            text = await self.pt_cli.prompt(default=default, async_=True)
+
         return text
 
     def init_io(self):
@@ -516,7 +622,7 @@ class ZMQTerminalInteractiveShell(SingletonConfigurable):
             # We can't set the buffer here, because it will be reset just after
             # this. Adding a callable to pre_run_callables does what we need
             # after the buffer is reset.
-            s = cast_unicode_py2(self.next_input)
+            s = self.next_input
 
             def set_doc():
                 self.pt_cli.app.buffer.document = Document(s)
@@ -528,12 +634,12 @@ class ZMQTerminalInteractiveShell(SingletonConfigurable):
                 set_doc()
             self.next_input = None
 
-    def interact(self, display_banner=None):
+    async def interact(self, loop=None, display_banner=None):
         while self.keep_running:
             print('\n', end='')
 
             try:
-                code = self.prompt_for_code()
+                code = await self.prompt_for_code()
             except EOFError:
                 if (not self.confirm_exit) or \
                         ask_yes_no('Do you really want to exit ([y]/n)?', 'y', 'n'):
@@ -543,13 +649,31 @@ class ZMQTerminalInteractiveShell(SingletonConfigurable):
                 if code:
                     self.run_cell(code, store_history=True)
 
+    async def _main_task(self):
+        loop = asyncio.get_running_loop()
+        tasks = [asyncio.create_task(self.interact(loop=loop))]
+
+        if self.include_other_output:
+            # only poll the iopub channel asynchronously if we
+            # wish to include external content
+            tasks.append(asyncio.create_task(self.handle_external_iopub(loop=loop)))
+
+        _, pending = await asyncio.wait(tasks, return_when=asyncio.FIRST_COMPLETED)
+
+        for task in pending:
+            task.cancel()
+        try:
+            await asyncio.gather(*pending)
+        except asyncio.CancelledError:
+            pass
+
     def mainloop(self):
         self.keepkernel = not self.own_kernel
         # An extra layer of protection in case someone mashing Ctrl-C breaks
         # out of our internal code.
         while True:
             try:
-                self.interact()
+                asyncio.run(self._main_task())
                 break
             except KeyboardInterrupt:
                 print("\nKeyboardInterrupt escaped interact()\n")
@@ -583,8 +707,8 @@ class ZMQTerminalInteractiveShell(SingletonConfigurable):
             return
 
         # flush stale replies, which could have been ignored, due to missed heartbeats
-        while self.client.shell_channel.msg_ready():
-            self.client.shell_channel.get_msg()
+        while run_sync(self.client.shell_channel.msg_ready)():
+            run_sync(self.client.shell_channel.get_msg)()
         # execute takes 'hidden', which is the inverse of store_hist
         msg_id = self.client.execute(cell, not store_history)
 
@@ -617,7 +741,8 @@ class ZMQTerminalInteractiveShell(SingletonConfigurable):
     #-----------------
 
     def handle_execute_reply(self, msg_id, timeout=None):
-        msg = self.client.shell_channel.get_msg(block=False, timeout=timeout)
+        kwargs = {"timeout": timeout}
+        msg = run_sync(self.client.shell_channel.get_msg)(**kwargs)
         if msg["parent_header"].get("msg_id", None) == msg_id:
 
             self.handle_iopub(msg_id)
@@ -625,8 +750,8 @@ class ZMQTerminalInteractiveShell(SingletonConfigurable):
             content = msg["content"]
             status = content['status']
 
-            if status == 'aborted':
-                self.write('Aborted\n')
+            if status == "aborted":
+                sys.stdout.write("Aborted\n")
                 return
             elif status == 'ok':
                 # handle payloads
@@ -656,7 +781,8 @@ class ZMQTerminalInteractiveShell(SingletonConfigurable):
         ## Get the is_complete response:
         msg = None
         try:
-            msg = self.client.shell_channel.get_msg(block=True, timeout=timeout)
+            kwargs = {"timeout": timeout}
+            msg = run_sync(self.client.shell_channel.get_msg)(**kwargs)
         except Empty:
             warn('The kernel did not respond to an is_complete_request. '
                  'Setting `use_kernel_is_complete` to False.')
@@ -685,11 +811,9 @@ class ZMQTerminalInteractiveShell(SingletonConfigurable):
     include_other_output = Bool(False, config=True,
         help="""Whether to include output from clients
         other than this one sharing the same kernel.
-
-        Outputs are not displayed until enter is pressed.
         """
     )
-    other_output_prefix = Unicode("[remote] ", config=True,
+    other_output_prefix = Unicode("Remote ", config=True,
         help="""Prefix to add to outputs coming from clients other than this one.
 
         Only relevant if include_other_output is True.
@@ -712,6 +836,13 @@ class ZMQTerminalInteractiveShell(SingletonConfigurable):
         else:
             return from_here
 
+    async def handle_external_iopub(self, loop=None):
+        while self.keep_running:
+            # we need to check for keep_running from time to time
+            poll_result = await self.client.iopub_channel.socket.poll(500)
+            if poll_result:
+                self.handle_iopub()
+
     def handle_iopub(self, msg_id=''):
         """Process messages on the IOPub channel
 
@@ -720,10 +851,9 @@ class ZMQTerminalInteractiveShell(SingletonConfigurable):
 
            It only displays output that is caused by this session.
         """
-        while self.client.iopub_channel.msg_ready():
-            sub_msg = self.client.iopub_channel.get_msg()
+        while run_sync(self.client.iopub_channel.msg_ready)():
+            sub_msg = run_sync(self.client.iopub_channel.get_msg)()
             msg_type = sub_msg['header']['msg_type']
-            parent = sub_msg["parent_header"]
 
             # Update execution_count in case it changed in another session
             if msg_type == "execute_input":
@@ -732,6 +862,7 @@ class ZMQTerminalInteractiveShell(SingletonConfigurable):
             if self.include_output(sub_msg):
                 if msg_type == 'status':
                     self._execution_state = sub_msg["content"]["execution_state"]
+
                 elif msg_type == 'stream':
                     if sub_msg["content"]["name"] == "stdout":
                         if self._pending_clearoutput:
@@ -770,6 +901,12 @@ class ZMQTerminalInteractiveShell(SingletonConfigurable):
                         print()
                     print(text_repr)
 
+                    # Remote: add new prompt
+                    if not self.from_here(sub_msg):
+                        sys.stdout.write('\n')
+                        sys.stdout.flush()
+                        self.print_remote_prompt()
+
                 elif msg_type == 'display_data':
                     data = sub_msg["content"]["data"]
                     handled = self.handle_rich_data(data)
@@ -780,11 +917,19 @@ class ZMQTerminalInteractiveShell(SingletonConfigurable):
                         if 'text/plain' in data:
                             print(data['text/plain'])
 
+                # If execute input: print it
                 elif msg_type == 'execute_input':
                     content = sub_msg['content']
-                    if not self.from_here(sub_msg):
-                        sys.stdout.write(self.other_output_prefix)
-                    sys.stdout.write('In [{}]: '.format(content['execution_count']))
+                    ec = content.get('execution_count', self.execution_count - 1)
+
+                    # New line
+                    sys.stdout.write('\n')
+                    sys.stdout.flush()
+
+                    # With `Remote In [3]: `
+                    self.print_remote_prompt(ec=ec)
+
+                    # And the code
                     sys.stdout.write(content['code'] + '\n')
 
                 elif msg_type == 'clear_output':
@@ -823,33 +968,31 @@ class ZMQTerminalInteractiveShell(SingletonConfigurable):
             from PIL import Image, ImageShow
         except ImportError:
             return False
-        raw = base64.decodestring(data[mime].encode('ascii'))
+        raw = base64.decodebytes(data[mime].encode('ascii'))
         img = Image.open(BytesIO(raw))
         return ImageShow.show(img)
 
     def handle_image_stream(self, data, mime):
-        raw = base64.decodestring(data[mime].encode('ascii'))
+        raw = base64.decodebytes(data[mime].encode('ascii'))
         imageformat = self._imagemime[mime]
         fmt = dict(format=imageformat)
         args = [s.format(**fmt) for s in self.stream_image_handler]
-        with open(os.devnull, 'w') as devnull:
-            proc = subprocess.Popen(
-                args, stdin=subprocess.PIPE,
-                stdout=devnull, stderr=devnull)
+        with subprocess.Popen(args, stdin=subprocess.PIPE,
+                              stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL) as proc:
             proc.communicate(raw)
-        return (proc.returncode == 0)
+            return (proc.returncode == 0)
 
     def handle_image_tempfile(self, data, mime):
-        raw = base64.decodestring(data[mime].encode('ascii'))
+        raw = base64.decodebytes(data[mime].encode('ascii'))
         imageformat = self._imagemime[mime]
         filename = 'tmp.{0}'.format(imageformat)
-        with NamedFileInTemporaryDirectory(filename) as f, \
-                open(os.devnull, 'w') as devnull:
-            f.write(raw)
-            f.flush()
-            fmt = dict(file=f.name, format=imageformat)
+        with TemporaryDirectory() as tempdir:
+            fullpath = os.path.join(tempdir, filename)
+            with open(fullpath, 'wb') as f:
+                f.write(raw)
+            fmt = dict(file=fullpath, format=imageformat)
             args = [s.format(**fmt) for s in self.tempfile_image_handler]
-            rc = subprocess.call(args, stdout=devnull, stderr=devnull)
+            rc = subprocess.call(args, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
         return (rc == 0)
 
     def handle_image_callable(self, data, mime):
@@ -862,7 +1005,7 @@ class ZMQTerminalInteractiveShell(SingletonConfigurable):
     def handle_input_request(self, msg_id, timeout=0.1):
         """ Method to capture raw_input
         """
-        req = self.client.stdin_channel.get_msg(timeout=timeout)
+        req = run_sync(self.client.stdin_channel.get_msg)(timeout=timeout)
         # in case any iopub came while we were waiting:
         self.handle_iopub(msg_id)
         if msg_id == req["parent_header"].get("msg_id"):
@@ -891,6 +1034,6 @@ class ZMQTerminalInteractiveShell(SingletonConfigurable):
 
             # only send stdin reply if there *was not* another request
             # or execution finished while we were reading.
-            if not (self.client.stdin_channel.msg_ready() or
-                    self.client.shell_channel.msg_ready()):
+            if not (run_sync(self.client.stdin_channel.msg_ready)() or
+                    run_sync(self.client.shell_channel.msg_ready)()):
                 self.client.input(raw_data)
